def test_ccp_get_password():
    from pypas.central_credential_provider import CentralCredentialProvider
    from pypas.model.reqresuser import ReqresUser

    ccp = CentralCredentialProvider("https://reqres.in/")

    creds = ccp.credentials.get_credential("2")
    print(creds)
    assert isinstance(creds, ReqresUser)



import ipaddress


def verify_is_valid_ip_address(ip_address: str) -> bool:
    """Verify if the given IP address is valid.

    Args:
        ip_address (str): The IP address to be verified.

    Returns:
        bool: True if the IP address is valid, False otherwise.
    """
    try:
        ipaddress.ip_address(ip_address)
        return True
    except ValueError:
        return False


def validate_pem_format(pem: str) -> bool:
    """Validate if the given string is in PEM format.

    Args:
        pem (str): The string to be validated.

    Returns:
        bool: True if the string is in PEM format, False otherwise.
    """
    # Pure certificate without private key
    if pem.startswith("-----BEGIN CERTIFICATE-----") and pem.endswith("-----END CERTIFICATE-----"):
        return True
    # Certificate with private key
    elif pem.startswith("-----BEGIN PRIVATE KEY-----") and pem.endswith("-----END CERTIFICATE-----"):
        return True
    return False


def remove_none_values_from_dict(d: dict) -> dict:
    """Remove None values from a dictionary.

    Args:
        d (dict): The dictionary to be cleaned.

    Returns:
        dict: The cleaned dictionary.
    """
    return {k: v for k, v in d.items() if v is not None}



"""ApplicationAuthenticationMethod model class."""
from dataclasses import dataclass
from enum import Enum


class ApplicationAuthenticationMethodType(Enum):
    """ApplicationAuthenticationMethodType enum.
    Describes the type of authentication method set for an application.
    """

    machineAddress = 1
    osUser = 2
    path = 3
    hashValue = 4
    certificateattr = 5


@dataclass
class ApplicationAuthenticationMethod:
    """ApplicationAuthenticationMethod model class."""

    AppID: str
    AuthType: ApplicationAuthenticationMethodType
    AuthValue: str
    Comment: str
    IsFolder: str
    authId: int
    AllowInternalScripts: bool
    Subject: str
    Issuer: str
    SubjectAlternativeName: str



"""Group model."""
from dataclasses import dataclass
from enum import Enum
from typing import List


class GroupType(Enum):
    """Describes the type of group."""

    Vault = 1
    Directory = 2


@dataclass
class GroupMember:
    """Describes a member of a group."""

    id: int
    UserName: str


@dataclass
class Group:
    """Describe a group with its properties and members."""

    id: int
    groupType: GroupType
    groupName: str
    description: str
    location: str
    directory: str
    dn: str
    members: List[GroupMember]



VALID_IPV4_ADDRESS = "127.0.0.1"
VALID_IPV6_ADDRESS = "2001:0db8:85a3:0000:0000:8a2e:0370:7334"
INVALID_IPADDRESS = "abc"

VALID_PEM_CERTIFICATE = """-----BEGIN CERTIFICATE-----
MIIFaDCCBFCgAwIBAgISESHkvZFwK9Qz0KsXD3x8p44aMA0GCSqGSIb3DQEBCwUA
VQQDDBcqLmF3cy10ZXN0LnByb2dyZXNzLmNvbTCCASIwDQYJKoZIhvcNAQEBBQAD
ggEPADCCAQoCggEBAMGPTyynn77hqcYnjWsMwOZDzdhVFY93s2OJntMbuKTHn39B
bml6YXRpb252YWxzaGEyZzIuY3JsMIGgBggrBgEFBQcBAQSBkzCBkDBNBggrBgEF
BQcwAoZBaHR0cDovL3NlY3VyZS5nbG9iYWxzaWduLmNvbS9jYWNlcnQvZ3Nvcmdh
bml6YXRpb252YWxzaGEyZzJyMS5jcnQwPwYIKwYBBQUHMAGGM2h0dHA6Ly9vY3Nw
lffygD5IymCSuuDim4qB/9bh7oi37heJ4ObpBIzroPUOthbG4gv/5blW3Dc=
-----END CERTIFICATE-----"""
INVALID_PEM_CERTIFICATE = """---this is a random text---"""

TEST_DICT_WITH_NONES = {
    "key1": "value1",
    "key2": None,
    "key3": "value3",
    "key4": None,
}

TEST_DICT_WITHOUT_NONES = {
    "key1": "value1",
    "key3": "value3",
}


def test_verify_is_valid_ip_address():
    from pypas.utils import verify_is_valid_ip_address

    assert verify_is_valid_ip_address(VALID_IPV4_ADDRESS) is True
    assert verify_is_valid_ip_address(VALID_IPV6_ADDRESS) is True
    assert verify_is_valid_ip_address(INVALID_IPADDRESS) is False


def test_validate_pem_format():
    from pypas.utils import validate_pem_format

    assert validate_pem_format(VALID_PEM_CERTIFICATE) is True
    assert validate_pem_format(INVALID_PEM_CERTIFICATE) is False


def test_remove_none_values_from_dict():
    from pypas.utils import remove_none_values_from_dict

    assert remove_none_values_from_dict(TEST_DICT_WITH_NONES) == TEST_DICT_WITHOUT_NONES



from pypas.central_credential_provider import CentralCredentialProvider
from pypas.model.reqresuser import ReqresUser
from pypas.api.authentication_api import AuthMethod
import dataclasses

from pypas.vault import Vault

ccp = CentralCredentialProvider("https://reqres.in/")

creds = ccp.credentials.get_credential("ccp_appid", "ww_mysafe")

print(creds)
print(type(creds))
print(dataclasses.is_dataclass(creds))
for cred in creds:
    print(cred)
    print(type(cred))
    print(isinstance(cred, ReqresUser))


vault = Vault("https://reqres.in/")

vault.Authentication.logon("myuser", "mypassword", auth_method=AuthMethod.CyberArk)

safe = vault.Safes.get("ww_mysafe")



"""Safe model""" ""
from dataclasses import dataclass
from typing import List


@dataclass
class SafeCreator:
    """
    Describes the creator of a safe.

    Attributes:
        id (str): The unique ID of the creator.

        name (str): The unique name of the creator.
    """

    id: str
    name: str


@dataclass
class SafeAccount:
    """
    Describe a minimal representation of an account in a safe.

    Attributes:
        id (str): The unique ID of the account.

        name (str): The unique name of the account.
    """

    id: str
    name: str


@dataclass
class Safe:
    """
    Information about a specific safe in the vault.


    Attributes:
        safeUrlId (str): The unique ID of the Safe used when calling Safe APIs.

        safeName (str): The unique name of the Safe.

        safeNumber (int): The unique numerical ID of the Safe.

        description (str): The description of the Safe.

        location (str): The location of the Safe in the Vault.

        creator (SafeCreator): The safe's creator.

        olacEnabled (bool): Whether or not Object Level Access Control is enabled for the Safe.

        managingCPM (str): The name of the CPM user managing the Safe.

        numberOfVersionsRetention (int): The number of retained versions of every password that is stored in the Safe.

        numberOfDaysRetention (int): The number of days that password versions are saved in the Safe.

        autoPurgeEnabled (bool): Whether or not files will be automatically purged after the end of the
        Object History Retention Period defined in the Safe properties.
        Report Safes and PSM Recording Safes are created automatically with AutoPurgeEnabled set to Yes.
        These Safes cannot be managed by the CPM.

        creationTime (int): The Unix creation time of the Safe.

        lastModificationTime (int): The Unix time when the Safe was last updated.

        accounts (List[SafeAccount]): The safe's accounts.

        isExiredMember (bool): Whether or not the membership for the Safe is expired.
        For expired members, the value is True.
    """

    safeUrlId: str
    safeName: str
    safeNumber: int
    description: str
    location: str
    creator: SafeCreator
    olacEnabled: bool
    managingCPM: str
    numberOfVersionsRetention: int
    numberOfDaysRetention: int
    autoPurgeEnabled: bool
    creationTime: int
    lastModificationTime: int
    accounts: List[SafeAccount]
    isExiredMember: bool



from enum import Enum

from pypas.utils import remove_none_values_from_dict


class AuthMethod(Enum):
    """Authentication method enumeration"""

    CyberArk = 1
    Windows = 2
    LDAP = 3
    RADIUS = 4


class Authentication:
    """Authentication API endpoint"""

    def __init__(self, vault):
        self.vault = vault

    def logon(
        self,
        username: str,
        password: str,
        new_password: str = None,
        auth_method: AuthMethod = AuthMethod.CyberArk,
        concurrent_session: bool = True,
    ):
        """Logon to the vault.
        Relevant CyberArk Documentation:
        https://docs.cyberark.com/PAS/12.6/en/Content/SDK/CyberArk%20Authentication%20-%20Logon_v10.htm
        """

        request_url = f"{self.vault.base_url}PasswordVault/api/auth/{auth_method}/Logon/"

        data = {
            "username": username,
            "password": password,
            "newPassword": new_password,
            "concurrentSession": concurrent_session,
        }

        data = remove_none_values_from_dict(data)

        response = self.vault.session.post(request_url, json=data, verify=self.vault.verify_requests)

        response.raise_for_status()

        return response.json()



from typing import List
from pypas.model.reqresuser import ReqresUser
import urllib
import httpx

from pypas.utils import remove_none_values_from_dict


class Credentials:
    """Credentials endpoint of the Central Credential Provider."""

    def __init__(self, ccp):
        self.ccp = ccp

    def get_credential(
        self,
        app_id: str,
        safe: str,
        folder: str = None,
        object: str = None,
        user_name: str = None,
        address: str = None,
        database: str = None,
        policy_id: str = None,
        reason: str = None,
        connection_timeout: int = None,
        fail_on_password_change: bool = None,
        certificate_path: str = None,
        certificate_key_path: str = None,
        certificate_password: str = None,
    ) -> List[ReqresUser]:
        """Get a credential from a safe.
        Relevant CyberArk Documentation:
        https://docs.cyberark.com/AAM-CP/13.0/en/Content/CCP/Calling-the-Web-Service-using-REST.htm
        """
        params = {
            "AppID": app_id,
            "Safe": safe,
            "Folder": folder,
            "Object": object,
            "UserName": user_name,
            "Address": address,
            "Database": database,
            "PolicyID": policy_id,
            "Reason": reason,
            "ConnectionTimeout": connection_timeout,
            "FailOnPasswordChange": fail_on_password_change,
        }
        params = remove_none_values_from_dict(params)

        request_url = f"{self.ccp.ccp_base_url}{self.ccp.ccp_iis_site}/api/Accounts?{urllib.parse.urlencode(params)}"

        with self.ccp.get_session() as client:
            if not certificate_path:
                response = client.get(
                    request_url,
                )
            else:
                if certificate_path and certificate_key_path and certificate_password:
                    cert = (certificate_path, certificate_key_path, certificate_password)
                elif certificate_path and certificate_key_path:
                    cert = (certificate_path, certificate_key_path)
                else:
                    cert = certificate_path
                response = httpx.get(request_url, cert=cert)
            response.raise_for_status()

        users = []
        for user in response.json()["data"]:
            users.append(ReqresUser(**user))

        return users



"""SafeMember model class."""
from dataclasses import dataclass
from enum import Enum


class SafeMemberType(Enum):
    """Describes the type of a safe member."""

    User = 1
    Group = 2


@dataclass
class SafeMemberPermissions:
    """Describes the permissions of a safe member."""

    useAccounts: bool
    retrieveAccounts: bool
    listAccounts: bool
    addAccounts: bool
    updateAccountContent: bool
    updateAccountProperties: bool
    initiateCPMAccountManagementOperations: bool
    specifyNextAccountContent: bool
    renameAccounts: bool
    deleteAccounts: bool
    unlockAccounts: bool
    manageSafe: bool
    manageSafeMembers: bool
    backupSafe: bool
    viewAuditLog: bool
    viewSafeMembers: bool
    accessWithoutConfirmation: bool
    createFolders: bool
    deleteFolders: bool
    moveAccountsAndFolders: bool
    requestsAuthorizationLevel1: bool
    requestsAuthorizationLevel2: bool


@dataclass
class SafeMember:
    """Describes a safe member with their permissions."""

    safeUrlId: str
    safeName: str
    safeNumber: int
    memberId: int
    memberName: str
    memberType: SafeMemberType
    membershipExpirationDate: str
    isExpiredMembershipEnable: bool
    isPredefinedUser: bool
    permissions: SafeMemberPermissions
    isReadOnly: bool



from pypas.model.safe import Safe, SafeAccount, SafeCreator
from typing import List
import urllib

from pypas.utils import remove_none_values_from_dict


class Safes:
    """Safes API endpoint"""

    def __init__(self, vault):
        self.vault = vault

    def list(
        self,
        limt: int = None,
        offset: int = None,
        useCache: bool = False,
        sort: bool = False,
        search: str = None,
        includeAccounts: bool = False,
        extendedDetails: bool = False,
    ) -> List[Safe]:
        """List all safes.
        Relevant CyberArk Documentation:
        https://docs.cyberark.com/PAS/12.6/en/Content/SDK/Safes%20Web%20Services%20-%20List%20Safes.htm
        """

        params = {
            "limit": limt,
            "offset": offset,
            "useCache": useCache,
            "sort": sort,
            "search": search,
            "includeAccounts": includeAccounts,
            "extendedDetails": extendedDetails,
        }
        params = remove_none_values_from_dict(params)

        request_url = f"{self.vault.base_url}PasswordVault/API/Safes?{urllib.parse.urlencode(params)}"

        response = self.vault.get_request(request_url)

        safes = []
        for safe in response.json()["safes"]:
            safe_creator = SafeCreator(**response.json()["Creator"])
            accounts = []
            for account in response.json()["accounts"]:
                accounts.append(SafeAccount(**account))

            safe = Safe(
                safeUrlId=response.json()["safeUrlId"],
                safeName=response.json()["safeName"],
                safeNumber=response.json()["safeNumber"],
                description=response.json()["description"],
                location=response.json()["location"],
                creator=safe_creator,
                olacEnabled=response.json()["olacEnabled"],
                managingCPM=response.json()["managingCPM"],
                numberOfVersionsRetention=response.json()["numberOfVersionsRetention"],
                numberOfDaysRetention=response.json()["numberOfDaysRetention"],
                autoPurgeEnabled=response.json()["autoPurgeEnabled"],
                creationTime=response.json()["creationTime"],
                lastModificationTime=response.json()["lastModificationTime"],
                accounts=accounts,
                isExiredMember=response.json()["isExiredMember"],
            )
            safes.append(safe)
        return safes

    def get(self, safe_identifier: str) -> Safe:
        """Get a single safe by its name.

        Relevant CyberArk Documentation:
        https://docs.cyberark.com/PAS/12.6/en/Content/SDK/Safes%20Web%20Services%20-%20Get%20Safes%20Details.htm


        Args:
            safe_identifier (str): safeUrLId or safeName

        Returns:
            Safe: A safe object
        """

        request_url = f"{self.vault.base_url}PasswordVault/API/Safes/{safe_identifier}/"

        response = self.vault.get_request(request_url)
        response.raise_for_status()

        safe_creator = SafeCreator(**response.json()["Creator"])

        accounts = []
        for account in response.json()["accounts"]:
            accounts.append(SafeAccount(**account))

        return Safe(
            safeUrlId=response.json()["safeUrlId"],
            safeName=response.json()["safeName"],
            safeNumber=response.json()["safeNumber"],
            description=response.json()["description"],
            location=response.json()["location"],
            creator=safe_creator,
            olacEnabled=response.json()["olacEnabled"],
            managingCPM=response.json()["managingCPM"],
            numberOfVersionsRetention=response.json()["numberOfVersionsRetention"],
            numberOfDaysRetention=response.json()["numberOfDaysRetention"],
            autoPurgeEnabled=response.json()["autoPurgeEnabled"],
            creationTime=response.json()["creationTime"],
            lastModificationTime=response.json()["lastModificationTime"],
            accounts=accounts,
            isExiredMember=response.json()["isExiredMember"],
        )

    def create(
        self,
        name: str,
        description: str = None,
        location: str = None,
        number_of_versions_retention: int = None,
        number_of_days_retention: int = None,
        managing_cpm: str = None,
        olac_enabled: bool = False,
    ) -> Safe:
        """Create a new safe."""
        if number_of_days_retention and number_of_versions_retention:
            raise ValueError("Only one of number_of_days_retention and number_of_versions_retention can be set.")
        if not number_of_days_retention and not number_of_versions_retention:
            raise ValueError("Either number_of_days_retention or number_of_versions_retention must be set.")

        body = {
            "SafeName": name,
            "Description": description,
            "Location": location,
            "NumberOfVersionsRetention": number_of_versions_retention,
            "NumberOfDaysRetention": number_of_days_retention,
            "ManagingCPM": managing_cpm,
            "OlacEnabled": olac_enabled,
        }
        body = remove_none_values_from_dict(body)

        request_url = f"{self.vault.base_url}PasswordVault/API/Safes"

        response = self.vault.post_request(request_url, body=body)

        response.raise_for_status()

        safe_creator = SafeCreator(**response.json()["Creator"])

        accounts = []
        for account in response.json()["accounts"]:
            accounts.append(SafeAccount(**account))

        return Safe(
            safeUrlId=response.json()["safeUrlId"],
            safeName=response.json()["safeName"],
            safeNumber=response.json()["safeNumber"],
            description=response.json()["description"],
            location=response.json()["location"],
            creator=safe_creator,
            olacEnabled=response.json()["olacEnabled"],
            managingCPM=response.json()["managingCPM"],
            numberOfVersionsRetention=response.json()["numberOfVersionsRetention"],
            numberOfDaysRetention=response.json()["numberOfDaysRetention"],
            autoPurgeEnabled=response.json()["autoPurgeEnabled"],
            creationTime=response.json()["creationTime"],
            lastModificationTime=response.json()["lastModificationTime"],
            accounts=accounts,
            isExiredMember=response.json()["isExiredMember"],
        )



"""Application model."""
from dataclasses import dataclass


@dataclass
class Application:
    """Application model."""

    AccessPermittedFrom: str
    AccessPermittedTo: str
    AllowExtendedAuthenticationRestrictions: bool
    AppID: str
    BusinessOwnerEmail: str
    BusinessOwnerFName: str
    BusinessOwnerLName: str
    BusinessOwnerPhone: str
    Description: str
    Disabled: bool
    ExpirationDate: str
    Location: str



from dataclasses import dataclass
from .api.safe_api import Safes
from .api.authentication_api import Authentication


@dataclass
class Vault:
    """CyberArk Vault model class."""

    base_url: str
    verify_requests: bool = True

    def __post_init__(self):
        self.Safes = Safes(self)
        self.Authentication = Authentication(self)

    def get_request(self, url: str, params: dict = None) -> dict:
        """Make a GET request to the vault."""
        return {}

    def post_request(self, url: str, params: dict = None, body: dict = None) -> dict:
        """Make a POST request to the vault."""
        return {}

    def put_request(self, url: str, params: dict = None, body: dict = None) -> dict:
        """Make a PUT request to the vault."""
        return {}

    def delete_request(self, url: str, params: dict = None, body: dict = None) -> dict:
        """Make a DELETE request to the vault."""
        return {}



from dataclasses import dataclass
from .api.central_credential_provider_api import Credentials
import httpx
from httpx import Client


@dataclass
class CentralCredentialProvider:
    """CentralCredentialProvider model class."""

    ccp_base_url: str
    ccp_iis_site: str = "AIMWebService"
    ccp_verify_requests: bool = True
    session: httpx.Client = None

    def __post_init__(self):
        self.credentials = Credentials(self)

    def get_session(self) -> Client:
        """Get a session for the CCP."""
        if self.session:
            return self.session

        return Client(
            base_url=self.ccp_base_url,
            verify=self.ccp_verify_requests,
            headers={"Content-Type": "application/json"},
        )



"""User model."""
from dataclasses import dataclass
from enum import Enum
from typing import List

from pypas.model.group import GroupType


class UserSource(Enum):
    """Describes the source of a user."""

    CyberArk = 1
    LDAP = 2


class UserAuthenticationMethod(Enum):
    """Describes the authentication method available for a user."""

    AuthTypePass = 1
    AuthTypeRadius = 2
    AuthTypeLDAP = 3


class UserInterface(Enum):
    """Describes possible interfaces for a user."""

    PIMSU = 1
    PSM = 2
    PSMP = 3
    PVWA = 4
    WINCLIENT = 5
    PTA = 6
    PACLI = 7
    HTTPGW = 8
    EVD = 9
    PIMSu = 10
    AIMApp = 11
    CPM = 12
    PVWAApp = 13
    PSMApp = 14
    AppPrv = 15
    PSMPApp = 17


@dataclass
class UserBusinessAddress:
    """Describes the business address of a user."""

    workStreet: str
    workCity: str
    workState: str
    workZip: str
    workCountry: str


@dataclass
class UserInternet:
    """Describes the internet presence of a user."""

    homePage: str
    homeEmail: str
    businessEmail: str
    otherEmail: str


@dataclass
class UserPhones:
    """Describes the phone numbers of a user."""

    homeNumber: str
    businessNumber: str
    cellularNumber: str
    faxNumber: str
    pagerNumber: str


@dataclass
class UserPersonalDetails:
    """Describes the personal details of a user."""

    firstName: str
    middleName: str
    lastName: str
    street: str
    city: str
    state: str
    zip: str
    country: str
    title: str
    organization: str
    department: str
    profession: str


@dataclass
class UserGroupsMembership:
    """Describes a group the user is a member of."""

    groupID: int
    groupName: str
    groupType: GroupType


@dataclass
class User:
    """Describes a user with their properties."""

    id: str
    username: str
    source: UserSource
    userType: str
    componentUser: bool
    vaultAuthorization: List[str]
    location: str
    enableUser: bool
    changePassOnNextLogon: bool
    expiryDate: int
    suspended: bool
    lastSuccessfulLoginDate: int
    unAuthorizedInterfaces: List[UserInterface]
    authenticationMethod: List[UserAuthenticationMethod]
    passwordNeverExpires: bool
    distinguishedName: str
    description: str
    businessAddress: UserBusinessAddress
    internet: UserInternet
    phones: UserPhones
    personalDetails: UserPersonalDetails
    groupsMembership: List[UserGroupsMembership]
